from typing import Iterable

import tqdm
import torch.nn as nn
import torch.optim as optim

from torch.utils.tensorboard import SummaryWriter

from .utils.modules import ClassBundle


class Callback:

    def __init__(self) -> None:
        self._model: nn.Module = None
        self._optimizer: optim.Optimizer = None
        self._description: str = None

    def set_model(self, model: nn.Module) -> None:
        self._model = model

    def set_optimizer(self, optimizer: optim.Optimizer) -> None:
        self._optimizer = optimizer

    def set_params(self, params: dict) -> None:
        self._params = params

    def on_train_begin(self) -> None:
        pass

    def on_train_end(self) -> None:
        pass

    def on_epoch_begin(self, epoch: int) -> None:
        pass

    def on_epoch_end(self, epoch: int) -> None:
        pass

    def on_train_epoch_begin(self, epoch: int) -> None:
        pass

    def on_train_epoch_end(self, epoch: int, metrics: dict) -> None:
        pass

    def on_train_batch_begin(self, batch: int) -> None:
        pass

    def on_train_batch_end(self, batch: int, metrics: dict) -> None:
        pass

    def on_eval_begin(self, epoch: int = None) -> None:
        pass

    def on_eval_end(self, metrics: dict, epoch: int = None) -> None:
        pass

    def on_eval_batch_begin(self, batch: int) -> None:
        pass

    def on_eval_batch_end(self, batch: int, metrics: dict) -> None:
        pass


class Callbacks(ClassBundle[Callback]):

    def __init__(self,
                 callbacks: Iterable[Callback] = None) -> None:
        if callbacks is None:
            callbacks = []
        super().__init__(callbacks)


class TqdmTrainLogger(Callback):

    def __init__(self,
                 log_on: str = "step") -> None:
        super().__init__()

        if log_on == "step":
            self._log_on_step = True
        elif log_on == "epoch":
            self._log_on_step = False
        else:
            raise ValueError(f"Unknown log mode '{log_on}'")

        self._pbar = None
        self._description = None
        self._epochs = 1
        self._steps_per_epoch = 1

    def set_params(self, params: dict) -> None:
        self._epochs = params.get("epochs", 1)
        self._steps_per_epoch = params.get("steps_per_epoch", 1)
        self._description = params.get("description", None)

    def on_train_begin(self) -> None:
        if not self._log_on_step:
            self._pbar = tqdm.tqdm(total=self._epochs)
            self._pbar.set_description(self._description)

    def on_epoch_begin(self, epoch: int) -> None:
        if self._log_on_step:
            self._pbar = tqdm.tqdm(total=self._steps_per_epoch)
            self._pbar.set_description(f"Epoch {epoch}/{self._epochs}")

    def on_train_batch_end(self, batch: int, metrics: dict) -> None:
        if self._log_on_step:
            self._pbar.update(1)
            self._pbar.set_postfix_str(self._metrics_to_str(metrics))

    def on_train_epoch_end(self, epoch: int, metrics: dict) -> None:
        if not self._log_on_step:
            self._pbar.update(1)
            self._pbar.set_postfix_str(self._metrics_to_str(metrics))

    def on_eval_end(self, metrics: dict, epoch: int = None) -> None:
        self._pbar.set_postfix_str(
            f"{self._pbar.postfix}, {self._metrics_to_str(metrics, 'val')}")

    def on_epoch_end(self, epoch: int) -> None:
        if self._log_on_step:
            self._pbar.close()

    def on_train_end(self) -> None:
        if not self._log_on_step:
            self._pbar.close()

    @staticmethod
    def _metrics_to_str(metrics: dict, prefix: str = None) -> str:
        if prefix:
            return ", ".join([f"{prefix}_{k}={v:.7f}" for k, v in metrics.items()])
        else:
            return ", ".join([f"{k}={v:.7f}" for k, v in metrics.items()])


class TensorBoard(Callback):

    def __init__(self,
                 log_dir=None,
                 comment='',
                 purge_step=None,
                 max_queue=10,
                 flush_secs=120,
                 filename_suffix='',
                 **kwargs) -> None:
        super().__init__()
        self._sum_writter = SummaryWriter(
            log_dir=log_dir, comment=comment, purge_step=purge_step,
            max_queue=max_queue, flush_secs=flush_secs,
            filename_suffix=filename_suffix,
            **kwargs
        )

    def on_train_epoch_end(self, epoch: int, metrics: dict) -> None:
        for k, v in metrics.items():
            self._sum_writter.add_scalar(f"{k}/train", v, global_step=epoch)

    def on_eval_end(self, metrics: dict, epoch: int = None) -> None:
        for k, v in metrics.items():
            self._sum_writter.add_scalar(f"{k}/val", v, global_step=epoch)



from typing import Iterable, Tuple

import numpy as np

from torch.utils.data import Dataset, Subset


def fixed_split(dataset: Dataset, val_split, shuffle=True) -> Tuple[Subset, Subset]:
    assert(0 < val_split < 1)
    if shuffle:
        indices = np.random.permutation(len(dataset))
    else:
        indices = np.arange(len(dataset))
    split_idx = int(len(dataset) * val_split)
    return Subset(dataset, indices[split_idx:]), Subset(dataset, indices[:split_idx])


def k_fold_split(dataset: Dataset, k, shuffle=True) -> Iterable[Tuple[Subset, Subset]]:
    assert(k >= 2)
    if shuffle:
        indices = np.random.permutation(len(dataset))
    else:
        indices = np.arange(len(dataset))
    fold_size = (len(dataset) + k - 1) // k
    for i in range(k):
        val_indices = indices[i * fold_size: (i + 1) * fold_size]
        train_indices = np.concatenate(
            (indices[:i * fold_size], indices[(i + 1) * fold_size:]))
        yield Subset(dataset, train_indices), Subset(dataset, val_indices)



from functools import partial
from typing import Callable, Optional

import torch
import torch.nn as nn
import torchvision

from torchvision.models.regnet import BlockParams, SimpleStemIN, ResBottleneckBlock
from torchvision.models.regnet import model_urls

from .common import Model


__all__ = [
    "RegNet",
    "RegNetY400MF",
    "RegNetY800MF",
    "RegNetY1p6GF",
    "RegNetY3p2GF",
    "RegNetY8GF",
    "RegNetY16GF",
    "RegNetY32GF",
    "RegNetX400MF",
    "RegNetX800MF",
    "RegNetX1p6GF",
    "RegNetX3p2GF",
    "RegNetX8GF",
    "RegNetX16GF",
    "RegNetX32GF"
]


class RegNet(torchvision.models.RegNet, Model):

    ARCH: str = None

    def __init__(self,
                 block_params: BlockParams,
                 in_channels: int = 3,
                 num_classes: int = 1000,
                 stem_width: int = 32,
                 stem_type: Optional[Callable[..., nn.Module]] = None,
                 block_type: Optional[Callable[..., nn.Module]] = None,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation: Optional[Callable[..., nn.Module]] = None,
                 pretrained: bool = False) -> None:

        if stem_type is None:
            stem_type = SimpleStemIN
        if norm_layer is None:
            norm_layer = partial(nn.BatchNorm2d, eps=1e-05, momentum=0.1)
        if block_type is None:
            block_type = ResBottleneckBlock
        if activation is None:
            activation = nn.ReLU

        torchvision.models.RegNet.__init__(self,
                                           block_params,
                                           stem_width=stem_width,
                                           stem_type=stem_type,
                                           block_type=block_type,
                                           norm_layer=norm_layer,
                                           activation=activation)
        Model.__init__(self, pretrained=pretrained)

        if num_classes != self.fc.out_features:
            if pretrained:
                self._pretrained_modules.pop("fc")
            self.fc = torch.nn.Linear(self.fc.in_features, num_classes)
        if in_channels and in_channels != 3:
            if pretrained:
                raise ValueError(
                    f"Specifying input channels is not supported in pretrained model")
            self.stem = stem_type(in_channels, stem_width, norm_layer, activation)

    def _load_pretrained_weights(self) -> None:
        state_dict = torch.hub.load_state_dict_from_url(model_urls[self.ARCH])
        self.load_state_dict(state_dict)
        for state_key in state_dict:
            module_name = state_key.split(".")[0]
            if module_name not in self._pretrained_modules:
                self._pretrained_modules[module_name] = self.get_submodule(module_name)


class RegNetY400MF(RegNet):

    ARCH: str = "regnet_y_400mf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=16, w_0=48, w_a=27.89, w_m=2.09, group_width=8, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetY800MF(RegNet):

    ARCH: str = "regnet_y_800mf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=14, w_0=56, w_a=38.84, w_m=2.4, group_width=16, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetY1p6GF(RegNet):

    ARCH: str = "regnet_y_1_6gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=27, w_0=48, w_a=20.71, w_m=2.65, group_width=24, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetY3p2GF(RegNet):

    ARCH: str = "regnet_y_3_2gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=21, w_0=80, w_a=42.63, w_m=2.66, group_width=24, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetY8GF(RegNet):

    ARCH: str = "regnet_y_8gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=17, w_0=192, w_a=76.82, w_m=2.19, group_width=56, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)

class RegNetY16GF(RegNet):

    ARCH: str = "regnet_y_16gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=18, w_0=200, w_a=106.23, w_m=2.48, group_width=112, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)

class RegNetY32GF(RegNet):

    ARCH: str = "regnet_y_32gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=20, w_0=232, w_a=115.89, w_m=2.53, group_width=232, se_ratio=0.25, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetX400MF(RegNet):

    ARCH: str = "regnet_x_400mf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=22, w_0=24, w_a=24.48, w_m=2.54, group_width=16, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetX800MF(RegNet):

    ARCH: str = "regnet_x_800mf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=16, w_0=56, w_a=35.73, w_m=2.28, group_width=16, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetX1p6GF(RegNet):

    ARCH: str = "regnet_x_1_6gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=18, w_0=80, w_a=34.01, w_m=2.25, group_width=24, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetX3p2GF(RegNet):

    ARCH: str = "regnet_x_3_2gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=25, w_0=88, w_a=26.31, w_m=2.25, group_width=48, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


class RegNetX8GF(RegNet):

    ARCH: str = "regnet_x_8gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=23, w_0=80, w_a=49.56, w_m=2.88, group_width=120, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)

class RegNetX16GF(RegNet):

    ARCH: str = "regnet_x_16gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=22, w_0=216, w_a=55.59, w_m=2.1, group_width=128, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)

class RegNetX32GF(RegNet):

    ARCH: str = "regnet_x_32gf"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        block_params = BlockParams.from_init_params(
            depth=23, w_0=320, w_a=69.86, w_m=2.0, group_width=168, **kwargs)
        super().__init__(block_params,
                         in_channels=in_channels,
                         num_classes=num_classes,
                         pretrained=pretrained,
                         **kwargs)


import collections
from typing import OrderedDict


import torch.nn as nn


class Model:

    _pretrained_modules: OrderedDict[str, nn.Module]

    def __init__(self, pretrained: bool = False) -> None:
        self._pretrained_modules = collections.OrderedDict()
        if pretrained:
            self._load_pretrained_weights()

    def _load_pretrained_weights(self) -> None:
        raise NotImplementedError(
            f"Model '{self.__class__.__name__}' is not supported")

    @property
    def pretrained_modules(self) -> OrderedDict[str, nn.Module]:
        return self._pretrained_modules



from typing import Callable, Iterable, Optional, Union

import torch
import torch.nn as nn
import torch.optim as optim
import torchmetrics

from torch.utils.data import Dataset

from .metrics import Metrics, LossWrapper
from .callbacks import Callback, Callbacks, TqdmTrainLogger


class History(dict):

    def __init__(self) -> None:
        super().__init__()

    def _key(self, k: str, prefix: str) -> str:
        if prefix:
            k = f"{prefix}_{k}"
        return k

    def append(self, kv: dict, prefix: str = None) -> None:
        for k, v in kv.items():
            k = self._key(k, prefix)
            if k not in self:
                self[k] = []
            self[k].append(v)


class Trainer:

    def __init__(self,
                 model: nn.Module,
                 loss: nn.Module,
                 optimizer: optim.Optimizer,
                 metrics: Iterable[torchmetrics.Metric] = None,
                 device: Optional[Union[int, torch.device]] = None) -> None:

        self._model = model
        self._loss = loss
        self._optimizer = optimizer
        self._metrics = Metrics([LossWrapper(loss)])

        if metrics:
            self._metrics.append_all(metrics)

        self.device = device

        if device:
            self._apply(lambda t: t.to(device=device))

    def train(self,
              train_data: Dataset,
              batch_size: int = 32,
              epochs: int = 1,
              val_data: Dataset = None,
              callbacks: Iterable[Callback] = None,
              verbose: int = 1,
              description: str = None) -> History:
        # init history
        history = History()

        # init callbacks
        callbacks = self._init_callbacks(
            callbacks, verbose,
            epochs=epochs,
            steps_per_epoch=(len(train_data) + batch_size - 1) // batch_size,
            description=description
        )

        # train begin
        callbacks.on_train_begin()
        for e in range(1, epochs + 1):
            # epoch starts
            callbacks.on_epoch_begin(e)
            train_iter = torch.utils.data.DataLoader(train_data,
                                                     batch_size=batch_size, shuffle=True)
            # reset metrics and train on train batches
            callbacks.on_train_epoch_begin(e)
            self._metrics.reset()
            self._model.train()
            for b, (X, y) in enumerate(train_iter, 1):
                callbacks.on_train_batch_begin(b)
                if self.device:
                    X = X.to(device=self.device)
                    y = y.to(device=self.device)
                self._train_step(X, y)
                callbacks.on_train_batch_end(b, self._metrics.to_dict())
            callbacks.on_train_epoch_end(e, self._metrics.to_dict())
            history.append(self._metrics.to_dict())
            # reset metrics and validate on val batches
            if val_data is not None:
                callbacks.on_eval_begin(epoch=e)
                self._metrics.reset()
                self._model.eval()
                val_iter = torch.utils.data.DataLoader(
                    val_data, batch_size=batch_size)
                for b, (X, y) in enumerate(val_iter):
                    callbacks.on_eval_batch_begin(b)
                    if self.device:
                        X = X.to(device=self.device)
                        y = y.to(device=self.device)
                    self._eval_step(X, y)
                    callbacks.on_eval_batch_end(b, self._metrics.to_dict())
                callbacks.on_eval_end(self._metrics.to_dict(), epoch=e)
                history.append(self._metrics.to_dict(), prefix="val")
            # epoch ends
            callbacks.on_epoch_end(e)
        # train end
        callbacks.on_train_end()
        return history

    def _init_callbacks(self,
                        callbacks: Iterable[Callback],
                        verbose: int,
                        **params) -> Callbacks:
        callbacks = Callbacks(callbacks)
        if verbose == 1:
            callbacks.append(TqdmTrainLogger(log_on="step"))
        elif verbose == 2:
            callbacks.append(TqdmTrainLogger(log_on="epoch"))
        callbacks.set_model(self._model)
        callbacks.set_optimizer(self._optimizer)
        callbacks.set_params(params)
        return callbacks
    
    def _train_step(self, X: torch.Tensor, y: torch.Tensor):
        loss = self._metrics(self._model(X), y)[0]
        self._optimizer.zero_grad()
        loss.backward()
        self._optimizer.step()

    def _eval_step(self, X: torch.Tensor, y: torch.Tensor):
        with torch.no_grad():
            self._metrics.update(self._model(X), y)

    def _apply(self, fn: Callable):
        self._model = fn(self._model)
        self._loss = fn(self._loss)
        self._metrics = fn(self._metrics)
        return self

    def cuda(self, device: Optional[Union[int, torch.device]] = None):
        if isinstance(device, int):
            self.device = f"cuda:{device}"
        elif isinstance(device, torch.device):
            self.device = device
        else:
            self.device = "cuda"
        return self._apply(lambda t: t.cuda(device))

    def cpu(self):
        self.device = "cpu"
        return self._apply(lambda t: t.cpu())



import torch
import torchvision

from torchvision.models.vision_transformer import model_urls

from .common import Model


__all__ = [
    "ViT",
    "ViTB16",
    "ViTB32",
    "ViTL16",
    "ViTL32"
]


class ViT(torchvision.models.VisionTransformer, Model):

    ARCH: str = None

    def __init__(self, num_classes: int, pretrained: bool = False, **kwargs) -> None:
        image_size = kwargs.pop("image_size", 224)
        torchvision.models.VisionTransformer.__init__(self, 
                                                      image_size, 
                                                      num_classes=num_classes, 
                                                      **kwargs)
        Model.__init__(self, pretrained=pretrained)

    def _load_pretrained_weights(self) -> None:
        state_dict = torch.hub.load_state_dict_from_url(model_urls[self.ARCH])
        for state_key in list(state_dict.keys()):
            module_name = state_key.split(".")[0]
            if module_name == "class_token":
                continue
            if module_name == "heads" and (self.representation_size or self.num_classes != 1000):
                state_dict.pop(state_key)
                continue
            if module_name not in self._pretrained_modules:
                self._pretrained_modules[module_name] = self.get_submodule(module_name)
        self.load_state_dict(state_dict, strict=False)


class ViTB16(ViT):

    ARCH: str = "vit_b_16"

    def __init__(self, num_classes: int, pretrained: bool = False, **kwargs) -> None:
        super().__init__(num_classes,
                         patch_size=16,
                         num_layers=12,
                         num_heads=12,
                         hidden_dim=768,
                         mlp_dim=3072,
                         pretrained=pretrained,
                         **kwargs)


class ViTB32(ViT):

    ARCH: str = "vit_b_32"

    def __init__(self, num_classes: int, pretrained: bool = False, **kwargs) -> None:
        super().__init__(num_classes,
                         patch_size=32,
                         num_layers=12,
                         num_heads=12,
                         hidden_dim=768,
                         mlp_dim=3072,
                         pretrained=pretrained,
                         **kwargs)


class ViTL16(ViT):

    ARCH: str = "vit_l_16"

    def __init__(self, num_classes: int, pretrained: bool = False, **kwargs) -> None:
        super().__init__(num_classes,
                         patch_size=16,
                         num_layers=24,
                         num_heads=16,
                         hidden_dim=1024,
                         mlp_dim=4096,
                         pretrained=pretrained,
                         **kwargs)


class ViTL32(ViT):

    ARCH: str = "vit_l_32"

    def __init__(self, num_classes: int, pretrained: bool = False, **kwargs) -> None:
        super().__init__(num_classes,
                         patch_size=32,
                         num_layers=24,
                         num_heads=16,
                         hidden_dim=1024,
                         mlp_dim=4096,
                         pretrained=pretrained,
                         **kwargs)



from collections import OrderedDict
from typing import Callable, Iterable, List, Optional, Union

import torch
import torch.nn as nn
import torchmetrics

from .utils.modules import ClassBundle


class Metrics(ClassBundle[torchmetrics.Metric]):

    def __init__(self, 
                 metrics: Iterable[torchmetrics.Metric] = None) -> None:
        if metrics is None:
            metrics = []
        super().__init__(metrics)
        self._metrics = self._objs

    def __call__(self, preds: torch.Tensor, targets: torch.Tensor) -> List[torch.Tensor]:
        results = []
        for m in self._metrics:
            results.append(m(preds, targets))
        return results

    @property
    def names(self) -> List[str]:
        return list(map(lambda m: m._get_name().lower(), self._metrics))

    def reset(self):
        for m in self._metrics:
            m.reset()

    def update(self, preds: torch.Tensor, targets: torch.Tensor) -> None:
        for m in self._metrics:
            m.update(preds, targets)

    def to_dict(self) -> OrderedDict:
        results = OrderedDict()
        for m in self._metrics:
            results[m._get_name().lower()] = m.compute().detach().cpu()
        return results

    def _apply(self, fn: Callable):
        for i, m in enumerate(self._metrics):
            self._metrics[i] = fn(m)
        return self
    
    def cuda(self, device: Optional[Union[int, torch.device]] = None):
        return self._apply(lambda t: t.cuda(device))

    def cpu(self):
        return self._apply(lambda t: t.cpu())

    def to(self, device: Union[str, torch.device]):
        return self._apply(lambda t: t.to(device=device))


class LossWrapper(torchmetrics.Metric):

    def __init__(self, loss: nn.Module):
        super().__init__()
        self.add_state("loss", default=[], dist_reduce_fx="cat")
        self._loss_func = loss

    def update(self, preds: torch.Tensor, targets: torch.Tensor):
        self.loss.append(self._loss_func(preds, targets))

    def compute(self) -> torch.Tensor:
        if len(self.loss) == 1:
            return self.loss[0].mean()
        else:
            return torch.cat(self.loss).mean()

    def _get_name(self):
        return "Loss"



import torch
import torchvision

from torchvision.models.resnet import model_urls

from .common import Model


__all__ = [
    "ResNet",
    "ResNet10",
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNet152"
]


class ResNet(torchvision.models.ResNet, Model):

    ARCH: str = None

    def __init__(self, num_classes: int, in_channels: int = None, pretrained: bool = False, **kwargs) -> None:
        torchvision.models.ResNet.__init__(self, **kwargs)
        Model.__init__(self, pretrained=pretrained)

        if num_classes != self.fc.out_features:
            if pretrained:
                self._pretrained_modules.pop("fc")
            self.fc = torch.nn.Linear(self.fc.in_features, num_classes)
        if in_channels and in_channels != self.conv1.in_channels:
            if pretrained:
                raise ValueError(
                    f"Specifying input channels is not supported in pretrained model")
            self.conv1 = torch.nn.Conv2d(
                in_channels, self.conv1.out_channels, kernel_size=7, stride=2, padding=3, bias=False)

    def _load_pretrained_weights(self) -> None:
        state_dict = torch.hub.load_state_dict_from_url(model_urls[self.ARCH])
        self.load_state_dict(state_dict)
        for state_key in state_dict:
            module_name = state_key.split(".")[0]
            if module_name not in self._pretrained_modules:
                self._pretrained_modules[module_name] = self.get_submodule(module_name)


class ResNet10(ResNet):

    ARCH: str = "resnet10"

    def __init__(self, num_classes: int, in_channels: int = None) -> None:
        super().__init__(num_classes,
                         in_channels=in_channels,
                         block=torchvision.models.resnet.BasicBlock,
                         layers=[1, 1, 1, 1])


class ResNet18(ResNet):

    ARCH: str = "resnet18"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained=False) -> None:
        super().__init__(num_classes,
                         in_channels=in_channels,
                         block=torchvision.models.resnet.BasicBlock,
                         layers=[2, 2, 2, 2],
                         pretrained=pretrained)


class ResNet34(ResNet):

    ARCH: str = "resnet34"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained=False) -> None:
        super().__init__(num_classes,
                         in_channels=in_channels,
                         block=torchvision.models.resnet.BasicBlock,
                         layers=[3, 4, 6, 3],
                         pretrained=pretrained)


class ResNet50(ResNet):

    ARCH: str = "resnet50"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained=False) -> None:
        super().__init__(num_classes,
                         in_channels=in_channels,
                         block=torchvision.models.resnet.Bottleneck,
                         layers=[3, 4, 6, 3],
                         pretrained=pretrained)


class ResNet101(ResNet):

    ARCH: str = "resnet101"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained=False) -> None:
        super().__init__(num_classes,
                         in_channels=in_channels,
                         block=torchvision.models.resnet.Bottleneck,
                         layers=[3, 4, 23, 3],
                         pretrained=pretrained)


class ResNet152(ResNet):

    ARCH: str = "resnet152"

    def __init__(self, num_classes: int, in_channels: int = None, pretrained=False) -> None:
        super().__init__(num_classes,
                         in_channels=in_channels,
                         block=torchvision.models.resnet.Bottleneck,
                         layers=[3, 8, 36, 3],
                         pretrained=pretrained)





from typing import Any, Callable, Generic, Iterable, List, TypeVar, Union

T = TypeVar("T")


class ClassBundle(Generic[T]):

    def __init__(self, objs: Iterable[T]) -> None:
        self._objs: List[T] = list(objs)

    def __passthrough__(self, __name: str) -> Union[List, Callable]:
        def func(*args, **kwargs):
            results = []
            for _o in self._objs:
                results.append(getattr(_o, __name)(*args, **kwargs))
            return results
        if all(callable(getattr(_o, __name)) for _o in self._objs):
            return func
        else:
            return [getattr(_o, __name) for _o in self._objs]

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError as e:
            if not all(getattr(_o, __name, None) for _o in self._objs):
                raise(e)
            return self.__passthrough__(__name)

    def append(self, obj: T) -> None:
        self._objs.append(obj)

    def append_all(self, obj_list: Iterable[T]) -> None:
        for o in obj_list:
            self._objs.append(o)
